//
//  ComposeTextView.swift
//  Tusker
//
//  Created by Shadowfacts on 8/18/20.
//  Copyright © 2020 Shadowfacts. All rights reserved.
//

import SwiftUI

struct ComposeTextView: View {
    @Binding private var text: String
    private let placeholder: Text?
    private let minHeight: CGFloat
    
    private var heightDidChange: ((CGFloat) -> Void)?
    private var backgroundColor = UIColor.secondarySystemBackground
    
    @State private var height: CGFloat?
    
    init(text: Binding<String>, placeholder: Text?, minHeight: CGFloat = 150) {
        self._text = text
        self.placeholder = placeholder
        self.minHeight = minHeight
    }
    
    var body: some View {
        ZStack(alignment: .topLeading) {
            Color(backgroundColor)
            
            if text.isEmpty, let placeholder = placeholder {
                placeholder
                    .font(.body)
                    .foregroundColor(.secondary)
                    .offset(x: 4, y: 8)
            }
            
            WrappedTextView(
                text: $text,
                textDidChange: self.textDidChange,
                font: .preferredFont(forTextStyle: .body)
            )
                .frame(height: height ?? minHeight)
        }
    }
    
    private func textDidChange(textView: UITextView) {
        height = max(minHeight, textView.contentSize.height)
        heightDidChange?(height!)
    }
    
    func heightDidChange(_ callback: @escaping (CGFloat) -> Void) -> Self {
        var copy = self
        copy.heightDidChange = callback
        return copy
    }
    
    func backgroundColor(_ color: UIColor) -> Self {
        var copy = self
        copy.backgroundColor = color
        return copy
    }
}

struct WrappedTextView: UIViewRepresentable {
    typealias UIViewType = UITextView
    
    @Binding var text: String
    var textDidChange: ((UITextView) -> Void)?
    var font = UIFont.systemFont(ofSize: 20)
    
    @Environment(\.isEnabled) private var isEnabled: Bool
    
    func makeUIView(context: Context) -> UITextView {
        let textView = UITextView()
        textView.delegate = context.coordinator
        textView.isEditable = true
        textView.backgroundColor = .clear
        textView.font = font
        textView.adjustsFontForContentSizeCategory = true
        textView.textContainer.lineBreakMode = .byWordWrapping
        return textView
    }
    
    func updateUIView(_ uiView: UITextView, context: Context) {
        uiView.text = text
        uiView.isEditable = isEnabled
        context.coordinator.textView = uiView
        context.coordinator.text = $text
        context.coordinator.didChange = textDidChange
        // wait until the next runloop iteration so that SwiftUI view updates have finished and
        // the text view knows its new content size
        DispatchQueue.main.async {
            self.textDidChange?(uiView)
        }
    }
    
    func makeCoordinator() -> Coordinator {
        return Coordinator(text: $text, didChange: textDidChange)
    }
    
    class Coordinator: NSObject, UITextViewDelegate, ComposeTextViewCaretScrolling {
        weak var textView: UITextView?
        var text: Binding<String>
        var didChange: ((UITextView) -> Void)?
        var caretScrollPositionAnimator: UIViewPropertyAnimator?
        
        init(text: Binding<String>, didChange: ((UITextView) -> Void)?) {
            self.text = text
            self.didChange = didChange
            
            super.init()

            NotificationCenter.default.addObserver(self, selector: #selector(keyboardDidShow), name: UIResponder.keyboardDidShowNotification, object: nil)
        }
        
        @objc private func keyboardDidShow() {
            guard let textView,
                  textView.isFirstResponder else { return }
            ensureCursorVisible(textView: textView)
        }
        
        func textViewDidChange(_ textView: UITextView) {
            text.wrappedValue = textView.text
            didChange?(textView)
            
            ensureCursorVisible(textView: textView)
        }
    }
}


//struct ComposeTextView_Previews: PreviewProvider {
//    static var previews: some View {
//        ComposeTextView()
//    }
//}
